#!/usr/bin/perl
# -*- mode: Perl -*-
# digipass.pl
#
# Command line script for manipulating Digipass token data in an SQL database
# intended for integration with Radiator etc.
#
# Author: Mike McCauley (mikem@open.com.au)
# Copyright (C) 2004 Open System Consultants
# $Id: digipass.pl,v 1.4 2005/08/25 23:04:20 mikem Exp $

require "newgetopt.pl";
require Authen::Digipass;
use DBI;
require strict;

my @options =
    (
     'h',                   # Print usage
     'v',                   # Verbose
     'd',                   # Debug
     'n',                   # Dont actually update the database
     'dbsource=s',
     'dbusername=s',
     'dbauth=s',
     'importkey=s',         # Key for importing tokens from a file
     'app=s',               # The name of the Digipass application to use. Spaces are important
     'f',                   # Force updates
     'syncwindow=n',
     'table=s',
     );

&NGetOpt(@options) || &usage;
&usage if $opt_h;

# Default values for command line options
my $dbsource   = $main::opt_dbsource   || 'dbi:mysql:radius';
my $dbusername = $main::opt_dbusername || 'mikem';
my $dbauth     = $main::opt_dbauth     || 'fred';
my $importkey  = $main::opt_importkey  || '11111111111111111111111111111111'; # For importing
my $app        = $main::opt_app        || 'APPL 1      ';
my $syncwindow = $main::opt_syncwindow || 6;
my $table      = $main::opt_table      || 'TBL_VASCODP';

# First make sure we can connect to the database
# Open the database
print "Connecting to database with $dbsource, $dbusername, $dbauth\n" if $main::opt_v;
$dbh = DBI->connect($dbsource, $dbusername, $dbauth, 
		    {PrintError => 0
		     }) 
    || die "Could not connect to $dbsource: $DBI::errstr";

# Intialise the kernel paramters that Authen::Digipass requires for most calls.
$kparms = Authen::Digipass::KernelParms->new(SyncWindow => $syncwindow);

# Now find out the command they want to do
my $command = shift @ARGV;
if ($command eq 'import')
{
    import();
}
elsif ($command eq 'list')
{
    listAll();
}
elsif ($command eq 'reset')
{
    resetData();
}
elsif ($command eq 'resetstaticpassword')
{
    resetStaticPassword();
}
elsif ($command eq 'changestaticpassword')
{
    changeStaticPassword();
}
elsif ($command eq 'verify')
{
    verify();
}
elsif ($command eq 'assign')
{
    assign();
}
elsif ($command eq 'unlock')
{
    unlock();
}
elsif ($command eq 'info')
{
    info();
}
else
{
    usage();
}

sub usage
{
    print "usage: $0 [-h] [-d] [-v] [-n]
  [-dbsource dbi:db:etc] [-dbusername username] [-dbauth password]
  [-importkey nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn] [-app appname]
  [-syncwindow hours]
  command arg arg ...
  Commands:
    import file file ...
    list
    reset serialno serialno .... 
    resetstaticpassword serialno serialno .... 
    changestaticpassword serialno newpin
    verify serialno password
    assign serialno [username]
    unlock serialno tokencode
    info serialno serialno ....
  \n";
    exit;
}


#####################################################################
# Import all the tokens from a number of files in DPX format
sub import
{
    usage() unless @ARGV; # Need filenames

    foreach (@ARGV)
    {
	print "Starting import of tokens from from file '$_' with key $importkey\n" if $main::opt_v;
	my ($ret, $handle, $appl_count, $appl_names, $token_count);
	my $ret = Authen::Digipass::AAL2DPXInit($handle, $_, $importkey,  $appl_count, $appl_names, $token_count);
	if ($ret != 0)
	{
	    # Open failed
	    my $err = Authen::Digipass::AAL2DPXGetErrorMsg($ret);
	    print STDERR "Error with init file '$_': $err\n";
	    next;
	}
	print "found $token_count tokens for $appl_count applications ($appl_names)\n", if $main::opt_v;

	# Find out which app to import
	my @apps = unpack('a12 a12 a12 a12 a12 a12 a12 a12', $appl_names); # up to 8 x 12 byte app names
	splice(@apps, $appl_count);
	my $easyappnames = join(', ', map {"'$_'"} @apps);
	my $thisapp;
	if ($appl_count == 1)
	{
	    $thisapp = $apps[0];
	}
	else
	{
	    # Choose the app based on the default app name or the -app argument
	    my $i;
	    for ($i = 0; $i < $appl_count; $i++)
	    {
		if ($apps[$i] eq $app)
		{
		    # This is the app to import
		    $thisapp = $app;
		}
	    }
	    if (!defined $thisapp)
	    {
		print STDERR "Application $app not available for importing from this token file.\nThe foollowing applicaitons are available: $easyappnames.\nUse -app to specify which application to import\n";
		next;
	    }
	}

	my ($i, $serial, $type, $mode, $data);
	for ($i = 0; $i < $token_count; $i++)
	{
	    $ret = Authen::Digipass::AAL2DPXGetToken($handle, $kparms, $thisapp, $serial, $type, $mode, $data);
	    if ($ret < 0)
	    {
		print STDERR "Could not get token for application '$thisapp': " . Authen::Digipass::AAL2DPXGetErrorMsg($ret) . "\n";
		next;
	    }
	    print "Read token $serial, $type, $mode, $data\n" if $main::opt_v;
	    
	    # Now try to insert into the database
	    my $q = "insert into $table (DIGIPASS, DP_TYPE, ALGO_TYPE, DP_DATA) values ('$serial', '$type', '$mode', '$data')";
	    print "Running insert query: $q\n" if $main::opt_v;
	    if (!$dbh->do($q))
	    {
		my $err = $dbh->errstr;
		
		# Insert failed
		if ($main::opt_f)
		{
		    # Maybe force an update if there is already an entry for this digipass
		    $q = "select DIGIPASS from $table where DIGIPASS = '$serial'";
		    print "Insert failed but -f option is on, looking for existing record to update with $q\n" if $main::opt_v;
		    my ($dummy) = $dbh->selectrow_array($q);
		    if (defined $dummy)
		    {
			$q = "update $table set DP_TYPE='$type', ALGO_TYPE='$mode', DP_DATA='$data' where DIGIPASS = '$serial'";
			print "Updating existing record with: $q\n", if $main::opt_v;
			next if $main::opt_n;
			if (!$dbh->do($q))
			{
			    # The update failed too, bomb out
			    $err = $dbh->errstr;
			    print STDERR "update of database record for '$serial' failed: $err\n";
			    next;
			}
		    }
		    else
		    {
			print STDERR "insert into database for Digipass '$serial' failed: $err\n";
			next;
		    }
		}
		else
		{
		    print STDERR "insert into database for Digipass '$serial' failed: $err\n";
		    next;
		}
	    }
	    print "imported '$serial'\n";
	}
    }
}

#####################################################################
# Reset tokens
sub resetData
{
    usage() unless @ARGV; # Need serial

    foreach (@ARGV)
    {
	my $serial = normaliseSerial($_);
	my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
	print "Looking for Digipass to update: $q\n" if $main::opt_v;
	my ($data) = $dbh->selectrow_array($q);
	if (defined $data)
	{
	    my $ret = Authen::Digipass::AAL2ResetTokenInfo($data, $kparms);
	    if ($ret != 0)
	    {
		my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
		print STDERR "Could not ResetTokenInfo for Digipass $serial: $err\n";
	    }
	    else
	    {
		# Write the updated data baack to the database
		$q = "update $table set DP_DATA = '$data' where DIGIPASS = '$serial'";
		print "Running update query: $q\n", if $main::opt_v;
		next if $main::opt_n;
		if (!$dbh->do($q))
		{
		    # The update failed too, bomb out
		    $err = $dbh->errstr;
		    print STDERR "update of database record for '$serial' failed: $err\n";
		}
	    }
	}
	else
	{
	    print STDERR "Could not reset '$serial': no such record in database\n";
	}
    }
}

#####################################################################
# Reset a tokens static password
sub resetStaticPassword
{
    usage() unless @ARGV; # Need serial

    foreach (@ARGV)
    {
	my $serial = normaliseSerial($_);
	my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
	print "Looking for Digipass to update: $q\n" if $main::opt_v;
	my ($data) = $dbh->selectrow_array($q);
	if (defined $data)
	{
	    my $ret = Authen::Digipass::AAL2ResetStaticPassword($data, $kparms);
	    if ($ret != 0)
	    {
		my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
		print STDERR "Could not ResetStaticPassword for Digipass $serial: $err\n";
	    }
	    else
	    {
		# Write the updated data baack to the database
		$q = "update $table set DP_DATA = '$data' where DIGIPASS = '$serial'";
		print "Running update query: $q\n", if $main::opt_v;
		next if $main::opt_n;
		if (!$dbh->do($q))
		{
		    # The update failed too, bomb out
		    $err = $dbh->errstr;
		    print STDERR "update of database record for '$serial' failed: $err\n";
		}
	    }
	}
	else
	{
	    print STDERR "Could not reset '$serial': no such record in database\n";
	}
    }
}

#####################################################################
# Reset a tokens static password
sub changeStaticPassword
{
    usage() unless @ARGV >= 2; # Need serial and new password

    my $serial = normaliseSerial($ARGV[0]);
    my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
    print "Looking for Digipass to update: $q\n" if $main::opt_v;
    my ($data) = $dbh->selectrow_array($q);
    if (defined $data)
    {
	my $ret = Authen::Digipass::AAL2ChangeStaticPassword($data, $kparms, $ARGV[1], $ARGV[1]);
	if ($ret != 0)
	{
	    my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
	    print STDERR "Could not ResetStaticPassword for Digipass $serial: $err\n";
	}
	else
	{
	    # Write the updated data baack to the database
	    $q = "update $table set DP_DATA = '$data' where DIGIPASS = '$serial'";
	    print "Running update query: $q\n", if $main::opt_v;
	    next if $main::opt_n;
	    if (!$dbh->do($q))
	    {
		# The update failed too, bomb out
		$err = $dbh->errstr;
		print STDERR "update of database record for '$serial' failed: $err\n";
	    }
	}
    }
    else
    {
	print STDERR "Could not reset '$serial': no such record in database\n";
    }
}

#####################################################################
# Reset get the unlock code for a locked token
sub unlock
{
    usage() unless @ARGV >= 2; # Need serial and new password

    my $serial = normaliseSerial($ARGV[0]);
    my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
    print "Looking for Digipass to unlock: $q\n" if $main::opt_v;
    my ($data) = $dbh->selectrow_array($q);
    if (defined $data)
    {
	my $unlockcode;
	my $ret = Authen::Digipass::AAL2Unlock($data, $kparms, $ARGV[1], $unlockcode);
	if ($ret != 0)
	{
	    my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
	    print STDERR "Could not get Unlock code for Digipass $serial: $err\n";
	}
	else
	{
	    print "Unlock code for $ARGV[0]: $unlockcode\n";

	    # Write the updated data baack to the database
	    $q = "update $table set DP_DATA = '$data' where DIGIPASS = '$serial'";
	    print "Running update query: $q\n", if $main::opt_v;
	    next if $main::opt_n;
	    if (!$dbh->do($q))
	    {
		# The update failed too, bomb out
		$err = $dbh->errstr;
		print STDERR "update of database record for '$serial' failed: $err\n";
	    }
	}
    }
    else
    {
	print STDERR "Could not reset '$serial': no such record in database\n";
    }
}
#
####################################################################
# Verify password
sub verify
{
    usage() unless @ARGV >= 2; # Need serial and password

    my $serial = normaliseSerial($ARGV[0]);
    my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
    print "Looking for Digipass to update: $q\n" if $main::opt_v;
    my ($data) = $dbh->selectrow_array($q);
    if (defined $data)
    {
	my $ret = Authen::Digipass::AAL2VerifyPassword($data, $kparms, $ARGV[1], $ARGV[1]);
	if ($ret != 0)
	{
	    my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
	    print STDERR "Could not VerifyPassword for Digipass $serial: $err\n";
	}
	else
	{
	    # Write the updated data baack to the database
	    $q = "update $table set DP_DATA = '$data' where DIGIPASS = '$serial'";
	    print "Running update query: $q\n", if $main::opt_v;
	    next if $main::opt_n;
	    if (!$dbh->do($q))
	    {
		# The update failed too, bomb out
		$err = $dbh->errstr;
		print STDERR "update of database record for '$serial' failed: $err\n";
	    }
	    print "Password OK\n";
	}
    }
    else
    {
	print STDERR "Could not reset '$serial': no such record in database\n";
    }
}

#####################################################################
# List all tokens in the database
sub listAll
{
    my $q = "select DIGIPASS, USER_ID, DP_TYPE, ALGO_TYPE, DP_DATA from $table";
    my $sth =  $dbh->prepare($q);
    if (!$sth)
    {
	$err = $dbh->errstr;
	print STDERR "Could not prepare $q: $err\n";
	return;
    }
    if (!$sth->execute())
    {
	$err = $dbh->errstr;
	print STDERR "Could not execute $q: $err\n";
	return;
    }
    my @ret;
    while (@ret = $sth->fetchrow_array)
    {
	print join("\t", @ret) . "\n";
    }
}

#####################################################################
# Assign a token to a username
sub assign
{
    usage() unless @ARGV >= 1; # Need serial at least

    my $serial = normaliseSerial($ARGV[0]);
    my $q = "select DP_DATA, USER_ID from $table where DIGIPASS = '$serial'";
    print "Looking for Digipass to update: $q\n" if $main::opt_v;
    my ($dummy, $olduser) = $dbh->selectrow_array($q);
    if (defined $dummy)
    {
	my $name = $ARGV[1];
	my $qname = defined $name ? $dbh->quote($name) : 'NULL';

	# Deassign any other tokens already assigned to that name
	if (defined $ARGV[1])
	{
	    print "Running deassign query: $q\n", if $main::opt_v;
	    $q =  "update $table set USER_ID = NULL where USER_ID = $qname";
	    if (!$dbh->do($q))
	    {
		# The update failed too, bomb out
		$err = $dbh->errstr;
		print STDERR "Could not deassign other tokens already assigned to $name: $err\n";
	    }
	}

	# Write the updated data back to the database
	$q = "update $table set USER_ID = $qname where DIGIPASS = '$serial'";
	print "Running assign query: $q\n" if $main::opt_v;
	return if $main::opt_n;
	if (!$dbh->do($q))
	{
	    # The update failed too, bomb out
	    $err = $dbh->errstr;
	    print STDERR "update of database record for '$serial' failed: $err\n";
	}
    }
    else
    {
	print STDERR "Could not assign '$serial': no such record in database\n";
    }
}

#####################################################################
# Info about tokens
sub info
{
    usage() unless @ARGV; # Need serial

    foreach (@ARGV)
    {
	my $serial = normaliseSerial($_);
	my $q = "select DP_DATA from $table where DIGIPASS = '$serial'";
	print "Looking for Digipass to get info: $q\n" if $main::opt_v;
	my ($data) = $dbh->selectrow_array($q);
	if (defined $data)
	{
	    my %info;
	    my $ret = Authen::Digipass::AAL2GetTokenInfo($data, $kparms, \%info);
	    if ($ret != 0)
	    {
		my $err = Authen::Digipass::AAL2GetErrorMsg($ret);
		print STDERR "Could not GetTokenInfo for Digipass $serial: $err\n";
	    }
	    else
	    {
		print "Detailed information for Digipass $serial:\n";
		my $key;
		foreach $key (sort keys %info)
		{
		    print "$key: $info{$key}\n";
		}
	    }
	}
	else
	{
	    print STDERR "Could not get Digipass '$serial': no such record in database\n";
	}
    }
}

#####################################################################
# Take a serial number, append the default application name and a pad to 22 chars
sub normaliseSerial
{
    my ($s) = @_;

    $s .= $app unless length $s > 10;
    $s .= ' ' x 22;
    substr($s, 22) = '';
    return $s;
}
